;====================================================================
; NVIDIA NForce2 custom option ROM for DFI LanParty Ultra (A)
; vT4
; author: digitalbath
; 18 Nov 2021
; Credits to tictac, pinczakko, polygon, TzK and Infrared
; Version History
; vT4_ Test version
;====================================================================
use16                                           ; 16bit mode
ROM_SIZE_IN_BLOCK = 6                           ; 1 means ROM size is 1 block (512 bytes)
ROM_SIZE_IN_BYTE = ROM_SIZE_IN_BLOCK * 512      ; number of 512 byte blocks

define PCI_ADDR_PORT    0xCF8
define PCI_DATA_PORT    0xCFC
define CMOS_ADDR_PORT   0x70
define CMOS_DATA_PORT   0x71

macro configurePort pci
{
    mov eax, pci                                ; move address for 32bit register offset
    mov dx, PCI_ADDR_PORT                       ; pci register address port
    out dx, eax                                 ; set port we want to read
}

macro readPciReg pci
{
    configurePort pci                           ; pci register address port
    mov dx, PCI_DATA_PORT                       ; pci register data port
    in eax, dx                                  ; read register values
}

macro writePciReg pci, data, mask
{
    readPciReg pci                              ; pci register data port
    and eax, mask                               ; mask for data to remove pre-set value
    or eax, data                                ; add custom value to old register data
    out dx, eax                                 ; write new custom register value
}

macro readCmosReg reg, mask
{
    mov al, reg                                 ; set CMOS index
    out CMOS_ADDR_PORT, al                      ; send register offset
    in al, CMOS_DATA_PORT                       ; read CMOS data
    and al, mask                                ; mask the data and zero unneeded bits
}

ROMStart:
    db 0x055, 0x0AA                             ; ROM Header 55,AA -> Bootable rom
    db (ROMEnd - ROMStart)/512                  ; ROM Size in 512 bytes
    jmp MAIN                                    ; jump to main routine
    db 0                                        ; checksum, to be filled in later
    times (256)-($-$$) db 0                     ; locate Main routine at 100h


MAIN:
    ; save all registers for later restore
    pushfd
    pusha
;
;====================================================================
;
; CMOS register and mask used in _EN_CODE.bin and system module:
;
; Cmos register 75h 0xF0h    Drive Strength  Bus: 0 Device: 0 Function: 4 Mem Controller
; Dimm 1 Drive Strength, Registers 7D, 81   8-bit
; Dimm 2 Drive Strength, Registers 67, 73   8-bit
; Dimm 3 Drive Strength, Registers 65, 71   8-bit
;
;
; Cmos register 75h 0x0Fh    Slew Rate Bus: 0 Device: 0 Function: 4 Mem Controller
; Dimm 1 Slew Rate, Register 7C 8-bit
; Dimm 2 Slew Rate, Register 66 8-bit
; Dimm 3 Slew Rate, Register 64 8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 11h
;   2: 22h
;   3: 33h
;   4: 44h
;   5: 55h
;   6: 66h
;   7: 77h
;   8: 88h
;   9: 99h
;  10: AAh
;  11: BBh
;  12: CCh
;  13: DDh
;  14: EEh
;  15: FFh
;
;
; Cmos register 76h 0x1Fh  T(rc) Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 90h ?-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1T
;   ...
;   1F: 0001.1111b = 31T
;
;
; Cmos register 77h 0x1Fh T(rfc)  Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 91h ?-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1T
;   ...
;   1F: 0001.1111b = 31T
;
;
; Cmos register 74h 0x1Fh    alpha timings  Bus: 0 Device: 0 Function: 1 Mem Controller
;  Register 97h F0h 8-bit T(DOE)
;  Register 97h 0Fh 8-bit T(RRD)
;  Register 96h F0h 8-bit T(W2P)
;  Register 96h 0Fh 8-bit T(W2R)
;  Register 95h F0h 8-bit T(REXT)
;  Register 95h 0Fh 8-bit T(R2P)
;  Register 94h F0h 8-bit T(TR2W)
;
;   0: Auto (keep value, skip custom code)
;   1: A 0-0-3-3-1-0-3: 97h:00 96h:33 95h:10 94h:30
;   2: B 1-3-3-3-2-1-3: 97h:13 96h:33 95h:21 94h:30
;   3: C 1-3-3-3-3-3-4: 97h:13 96h:33 95h:33 94h:40
;   4: D 1-3-4-3-3-4-4: 97h:13 96h:43 95h:34 94h:40
;   5: E 2-3-3-3-3-3-4: 97h:23 96h:33 95h:33 94h:40
;   6: F 2-3-5-3-2-2-3: 97h:23 96h:53 95h:22 94h:30
;   7: G 3-2-5-3-2-2-3: 97h:32 96h:53 95h:22 94h:30
;   8: H 3-4-4-4-3-4-5: 97h:34 96h:44 95h:34 94h:50
;   9: I 3-4-5-4-3-4-5: 97h:34 96h:54 95h:34 94h:50
;   A: J 4-4-5-4-3-3-5: 97h:44 96h:54 95h:33 94h:50
;   B: K 4-5-5-5-3-5-6: 97h:45 96h:55 95h:35 94h:60
;   C: L 5-4-6-3-3-5-6: 97h:54 96h:63 95h:35 94h:60
;   D: M 5-6-4-3-3-5-6: 97h:56 96h:43 95h:35 94h:60
;   E: N 5-5-6-5-3-5-6: 97h:55 96h:65 95h:35 94h:60
;   F: O 6-6-6-6-3-6-6: 97h:66 96h:66 95h:36 94h:60
; 
;
; Cmos register 78h 0x19h  T(ref)   Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 60h+61h 16-bit
;
;   00: Auto (keep value, skip custom code)
;   01: 0402 -> 0204h =  516 cycles
;   02: 8802 -> 0288h =  648 cycles
;   03: 0803 -> 0308h =  776 cycles
;   04: 0D03 -> 030Dh =  780 cycles
;   05: 8C03 -> 038Ch =  908 cycles
;   06: 0804 -> 0408h = 1032 cycles
;   07: 9004 -> 0490h = 1168 cycles
;   08: 1005 -> 0510h = 1296 cycles
;   09: 0006 -> 0600h = 1536 cycles
;   0A: 1006 -> 0610h = 1552 cycles
;   0B: 1806 -> 0618h = 1560 cycles
;   0C: 1807 -> 0718h = 1816 cycles
;   0D: 0008 -> 0800h = 2048 cycles
;   0E: 1008 -> 0810h = 2064 cycles
;   0F: 2009 -> 0920h = 2336 cycles
;   10: 000A -> 0A00h = 2560 cycles
;   11: 200A -> 0A20h = 2592 cycles
;   12: 000C -> 0C00h = 3072 cycles
;   13: 300C -> 0C30h = 3120 cycles
;   14: 300E -> 0E30h = 3632 cycles
;   15: 640E -> 0E64h = 3684 cycles
;   16: 2010 -> 1020h = 4128 cycles
;   17: 6410 -> 1064h = 4196 cycles
;   18: 4012 -> 1240h = 4672 cycles
;   19: 6412 -> 1264h = 4708 cycles
;
;
; Cmos register 73h 0x1Fh  romsips tables  Bus: 0 Device: 0 Function: 3 Mem Controller
;   8-bit Register 65h, 66h, 68h, 69h, 6Ah, 6Bh, 6Ch, 6Dh, 6Fh
;   0: Auto (keep value, skip custom code)
;   1: A DFI 12/10:  65h:06 66h:F7 68h:75 69h:00 6Ah:55 6Bh:00 6Ch:66 6Dh:55 6Fh:33
;   2: B DFI 12/18:  65h:06 66h:F7 68h:55 69h:00 6Ah:66 6Bh:00 6Ch:55 6Dh:55 6Fh:33
;   3: C DFI 12/31:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   4: D DFI 01/21:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:F0
;   5: E DFI 01/31:  65h:00 66h:F7 68h:33 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   6: F DFI 05/05:  65h:00 66h:F7 68h:55 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   7: G custom#1 :  65h:66 66h:F7 68h:22 69h:66 6Ah:55 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   8: H custom#2 :  65h:66 66h:77 68h:22 69h:07 6Ah:97 6Bh:00 6Ch:88 6Dh:88 6Fh:77
;   9: I custom#3 :  65h:44 66h:22 68h:22 69h:22 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:22
;   A: J custom#4 :  65h:44 66h:44 68h:44 69h:44 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:44
;   B: K custom#5 :  65h:66 66h:F7 68h:33 69h:66 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33   7Ch:44 48h:69 50h:B2
;   C: L custom#6 :  65h:66 66h:F7 68h:33 69h:66 6Ah:67 6Bh:00 6Ch:66 6Dh:66 6Fh:77   7Ch:44 48h:69 50h:B2
;   D:
;   E:
;   F:
;
;
;====================================================================
;
;
; TRC
;
    writePciReg 0x80000190, 0x0000000D, 0xFFFFFFE0     ; offset 90, Value 13 (0Dh)
;
; TRFC
;
    writePciReg 0x80000190, 0x00001000, 0xFFFFE0FF     ; offset 91, value 16 (10h)
;
; slew Rate
;
    writePciReg 0x80000464, 0x00880088, 0xFF00FF00     ; offset 64 and 66, value 88h
    writePciReg 0x8000047C, 0x00000088, 0xFFFFFF00     ; offset 7C, 88h
;
; drive strength
;
    writePciReg 0x80000464, 0x44004400, 0x00FF00FF     ; offset 65 and 67, value 44h
    writePciReg 0x80000470, 0x44004400, 0x00FF00FF     ; offset 71 and 73, value 44,
    writePciReg 0x8000047C, 0x00004400, 0xFFFF00FF     ; offset 7D, value 44h
    writePciReg 0x80000480, 0x00004400, 0xFFFF00FF     ; offset 81, value 44h
;
; alpha timings
;
    writePciReg 0x80000194, 0x34443450, 0x0000000F     ; 97h:34 96h:44 95h:34 94h:50
;
; TREF
;
    writePciReg 0x80000160, 0x00000A00, 0xFFFF0000     ; 2560 Cycles
;
; additional PCI register ind b0d0f3
;
    writePciReg 0x80000348, 0x00000066, 0xFFFFFF00     ; 48h:66
    writePciReg 0x8000034C, 0x00000090, 0xFFFFFF00     ; 4Ch:90
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ; 50h:B2 
    writePciReg 0x80000374, 0x00000062, 0xFFFFFF00     ; 74h:62 
;
    jmp codeend
;
;====================================================================

codeend:
    popa
    popfd
    retf                                        ; return far to system bios routine

                                        
  times (ROM_SIZE_IN_BYTE-$) db 0               ; use 00h as the padding bytes until we reach the ROM size
                                                ; The last byte (512th) will be the patch_byte for the checksum
                                                ; patch_byte is calculated and automagically inserted below
  PREV_CHKSUM = 0
  repeat $
  load CHKSUM byte from %-1
  CHKSUM = (PREV_CHKSUM + CHKSUM) mod 0x100
  PREV_CHKSUM = CHKSUM
  end repeat
  
  store byte (0x100 - CHKSUM) at ($-1)          ; store the patch_byte
ROMEnd:
